/* Thermostat ATtiny45                                          */
/* Work with VT100 terminal or terminal emulator                */
/* on 2400 baud 8 data bit 1 or 2 stop bit no parity            */
/* no flow control.                                             */
/* Command set is                                               */
/* ?<CR>         - print current thresholds values              */
/* TON=xxx<CR>   - Set On threshold temperature in Celsius      */
/* TOFF=xxx<CR>  - Set Off threshold temperature in Celsius     */
/*--------------------------------------------------------------*/
/* Pin (num)  Name   Description                                */
/* PB0 (5)  Out   Triac gate     		 						*/
/* PB1 (6)  Tx    UART Tx pin     			  					*/
/* PB2 (7)  Rx    UART Rx pin      			  					*/
/* PB3 (2)  NC  Tts    Analog input. 100k NTC connected from +5V */
/*                   to Tts pin and pulled down with 5.61k      */
/* PB4 (3)  NC         			               					*/
/* PB5 (1)  Reset   */ 



 #include <avr/io.h>           /* AVR header file */
 #include <avr/interrupt.h>
// #include <avr/wdt.h>
 #include <avr/eeprom.h>
 #define SetBit(Reg, Bit)    (Reg |= (1 << Bit))
 #define ClearBit(Reg, Bit)  (Reg &= ~(1 << Bit))
 #define ToggleBit(Reg, Bit) (Reg ^= (1 << Bit))
 #define TestBit(Reg, Bit)   (Reg & (1 << Bit))
 #define InBit(Reg, Bit)    ((Bit)|=  )

/* Version 1.01avr */

/* Configuration word */
typedef unsigned int word;
typedef unsigned char byte;
#define NL      "\n\r"       /* New line */

/* EEPROM init */
char ee_tOn __attribute__((section(".eeprom"))); 
char ee_tOff __attribute__((section(".eeprom")));
char ee_num __attribute__((section(".eeprom"))) ;

//#define TRx      GPIO3
//#define TTx      GPIO4
//#define Out      GPIO5
#define TRx      0x02
#define TTx      0x01
#define Out      0x00

void idle(void);

static  byte  ms; // bit was
static  byte tOn, tOff;
static  byte Tts;

/* interrupt service routine variables */
static volatile byte rx_smpl_cntr; /* recevier sample counter */
static volatile byte rx_state;     /* receiver state (bit number) */
static volatile byte tx_smpl_cntr; /* transmitter sample counter */
static volatile byte tx_state;     /* transmitter state (bit number) */
static volatile byte txd;          /* transmitted data */
static volatile byte tx_ok;         /* 1 - transmit in progress */
static volatile byte mOut;          /* Output mirror */
#define BUFSTR_SIZE 10             /* Command line size */
char str[BUFSTR_SIZE];             /* Command line */
byte cmd;                          /* Command */

volatile byte iidx;                /* interrupt string index */
volatile byte tidx;                /* terminal string index */

void tputch(byte c)
{
 while (tx_ok == 1) idle(); /* do idle if UART is busy */
 txd = c;
 tx_ok = 1;
}


ISR(TIM0_OVF_vect)
 {
   static volatile byte rx_1;         /* Rx One counter */
   static volatile byte rx_0;         /* Rx Zero counter */
   static volatile byte rrx;          /* receveing data */
   static byte _138us;                /* 1/(2400*3) */
//  if (TestBit(PORTB, 0x00)) {ClearBit(PORTB, 0x00);} else {SetBit(PORTB, 0x00);};
/* start bit */
    TCNT0 = 188;  /*121  Reload timer   255-(7761760/8/7200) 121-2400, 188-4800*/
    /* Timebase calculation */
    if (_138us++ == 7)
     {
      ms = 1;
      _138us = 0;
     }

    if (mOut == 1) SetBit(PORTB,Out);
    rx_smpl_cntr++;

    if (TestBit(PINB,TRx)) /* One detection */
     {
      if (rx_1 < 255) rx_1++; /* One counter */
     }
    else rx_1 = 0;

    if (TestBit(PINB,TRx) == 0) /* Zero detection */
     {
      if (rx_0 < 255) rx_0++; /* Zero counter */
     }
    else rx_0 = 0;

    if (rx_smpl_cntr == 0)
     {
      rx_state = 0;   /* restart receiving */
      rx_1 = 0;
      rx_0 = 0;
      rrx = 0;
     }

    if (rx_state == 0)
     {
      rx_state = 1;   /* start receiving */
      rx_1 = 0;
      rx_0 = 0;
      rx_smpl_cntr = 0;
     }

    if ((rx_state == 1) && (rx_0 > 1) && (rx_smpl_cntr > 1))
     {
      rx_state = 2; /* start bit detected */
      rx_1 = 0;
      rx_0 = 0;
      rx_smpl_cntr = 0;
     }

    /* data bits receiving */
    if ((rx_state > 1) && (rx_state < 10) && (rx_smpl_cntr == 3))
     {
      rx_state++;
      rrx >>= 1;
      if (rx_1 > 1)
       {
        rrx |= 0x80;
       }
      rx_1 = 0;
      rx_0 = 0;
      rx_smpl_cntr = 0;
     }

    if ((rx_state == 10) && (rx_smpl_cntr == 3)) /* stop bit detection */
     {
      if (rx_1 > 1)
       {

        /* Received char validation and storing in to the string buffer */
        if (rrx > 'A') rrx &= 0x5f; /* Convert char to upper case */
        if (((rrx >= '0') && (rrx <= '9')) ||  /* is digit */
            ((rrx >= 'A') && (rrx <= 'Z')) ||  /* is alpha */
            (rrx == '=') || (rrx == '&') ||  /* is special */
            (rrx == ' ') || (rrx == '?') ||  /* is special */
            (rrx == '\r') || (rrx == '\b'))   /* is special */
         {
          str[iidx] = rrx;  /* Add char to the end of the command string */
          if (iidx < BUFSTR_SIZE-1) iidx++;
          else
           {
            iidx--;
            if (tidx > 0) tidx--;
           }
         }

       }
      rx_state = 0;
      rx_1 = 0;
      rx_0 = 0;
      rx_smpl_cntr = 0;
     }

    /* Transmit byte */
    if (tx_ok == 1)
     {
      /* Start bit */
      if ((tx_state == 0) && (tx_smpl_cntr == 1)) ClearBit(PORTB,TTx); //Tx

      /* data bits */
      if ((tx_state > 0) && (tx_state < 9) && (tx_smpl_cntr == 1))
       {
        if ((txd & 1) == 0) ClearBit(PORTB,TTx);
        else SetBit(PORTB,TTx);
        txd >>= 1;
       }

      /* first stop bit */
      if ((tx_state == 9) && (tx_smpl_cntr == 1)) SetBit(PORTB,TTx);

      /* second stop bit */
      if ((tx_state == 11) && (tx_smpl_cntr == 1))
       {
         SetBit(PORTB,TTx);
         tx_state = 0;
         tx_smpl_cntr = 0;
         tx_ok = 0;
       }

      tx_smpl_cntr++;
      if (tx_smpl_cntr == 3)
       {
        tx_state++;
        tx_smpl_cntr = 0;
       }
     }
    else
     { /* keep Tx high after transmition */
      tx_state = 0;
      tx_smpl_cntr = 0;
      SetBit(PORTB,TTx);
     }
   ClearBit(PORTB,Out);
 
 } /* end of interrupt service routine */

/* NTC temperature sensor line approximation */
struct tT_tbl
{
 byte k1;
 byte k2;
};

static const struct tT_tbl T_table[] =
 {
  {168, 34},
  {160, 34},
  {152, 34},
  {144, 34},
  {136, 34},
  {127, 34},
  {119, 34},
  {111, 34},
  {102, 34},
  {94, 34},
  {85, 34},
  {77, 34},
  {69, 41},
  {58, 51},
  {45, 70},
  {29, 116}
 };

byte thermo(byte Vt_hs)
{
 byte i, k1, k2;

 if (Vt_hs > 256-10) return 19;
 i = Vt_hs>>4;
 k1 = T_table[i].k1;
 k2 = T_table[i].k2;
 return (byte)(k1 - ((word)Vt_hs-(i<<4))*(k2>>6));
// return (byte)(k1 - ((word)Vt_hs-i*16)*k2/64);

}


/* AD measurement and average value calculation (low pass digital filter) */

/* Main state machine */
 static volatile word wVTavg;
        
void idle(void)
{
 signed int temp;
 word ad_tmp;
// word ttt;
 
 // wdt_reset();              /* Clear watchdog timer */
 if (ms == 1)
  {
     // Clear ADIF and start single conversion
    ADCSRA |= (1 << ADIF) | (1<<ADSC);
    // wait for conversion done, ADIF flag active
    while(!(ADCSRA & (1 << ADIF)));

// this works  
   ad_tmp = ((word)ADCL  >>1);
   ad_tmp += ((word)ADCH  <<7);// + (ADCL >>1);
   temp = (signed int)(ad_tmp - wVTavg);

    temp = (signed int) temp >>4;
    wVTavg += (signed int)temp;
       Tts = thermo((word)wVTavg>>7);    
   
    
   if ((mOut == 0) && (Tts < tOn)) mOut = 1;
   if ((mOut == 1) && (Tts > tOff)) mOut = 0;
   ms = 0;
  }
}



/* command line interpretator */
enum commands {cmEmpty,
               cmTOn,
               cmTOff,
               cmHelp,
               cmError,
               cmd_num
              };

const char * const cmdstr[] =
 {
  "TON=",
  "TOFF=",
  "?"
 };

/* Put int to UART  with lead zero */
void printi(byte value)
{
  byte const digits[]= { 100, 10, 1 };
  byte ch;
  byte i=0;

  for (i=0; i<3; i++)
  {
    ch='0';
    while(value>=digits[i])
    {
      ch++;
      value-=digits[i];
    }
    tputch(ch);
  }
}


/* put string to the terminal  */
void tputstr(const char const * ptr)
{
   char  c,i=0;
   while (0 != (c = ptr[i++]))
// while (c = ptr[i++])
  tputch(c);
}

/* convert string to number */
word stoi(char * str)
{
 byte b = 0;
 char c;

 while ((c = *str++))
  if ((c >= '0') && (c <= '9'))
    b = b * 10 + c - '0';
 return b;
}


void docmd(void)
{
  byte cc;

  switch (cmd)
   {
    case cmTOn:
                   cc = stoi(str + tidx);
                   tOn = cc;
                        //ee_write(tOn, ee_tOn);
		   eeprom_write_byte (&ee_tOn,tOn);
                   break;
    case cmTOff:
                   cc = stoi(str + tidx);
                   tOff = cc;
                         //ee_write(tOff, ee_tOff);
		   eeprom_write_byte (&ee_tOff,tOff);
                   break;
    case cmHelp:
                   tputstr("\n\rTon="); printi(tOn);
                   tputstr("\n\rToff="); printi(tOff);
                   break;
   }
}

/* Command line parser */
char parce(void)
{
 byte i;
 char cc;

#ifdef _USE_AT_
 if ((str[0] == 'A') && (str[1] == 'T')) /* AT prefix */
  {
   if (str[2] == 0) return cmd = cmEmpty;
   for(cmd=0; cmd < cmd_num; cmd++)
    {
     i = 0; tidx = 2;
     while(((cc = cmdstr[cmd][i]) != 0) && (cc == str[tidx++])) i++; /* compare with command */
     if (cmdstr[cmd][i] == 0) return cmd += 1; /* return command */
    }
  }
#else  /* _USE_AT_ */
 if (str[0] == 0) return cmd = cmEmpty;
 for(cmd=0; cmd < cmd_num; cmd++)
  {
   i = 0; tidx = 0;
   while(((cc = cmdstr[cmd][i]) != 0) && (cc == str[tidx++])) i++; /* compare with command */
   if (cmdstr[cmd][i] == 0) return cmd += 1; /* return command */
  }
#endif /* _USE_AT_ */
 return cmd = cmError;
}

/* Command line editor and executor */
void term(void)
{
 char c;

 if (tidx < iidx)
  {
   c = str[tidx];
   if (c == '\r')
    {
     tputstr(NL);
     str[iidx] = 0;
     cmd = parce(); /* call parcer */
     docmd();       /* execute command */
     if (cmd == cmError) tputstr("ERROR"NL);
     else tputstr(NL"OK"NL);
     tidx = iidx = 0;
    }
   else
   if (c == '\b')      /* if backspace pressed */
     {
      if (tidx > 0) tidx--;
      iidx = tidx;
      tputstr("\b \b");
     }
   else
    {
     if (tidx == 0) tputstr(NL);  /* print echo from the new line */
     tputch(c);
     tidx++;
    }
  }
}

int main(void) /* Entry point */
{
//OSCCAL = _READ_OSCCAL_DATA();
/*-------------- Init ports ----------------------*/
DDRB =    0b00000011;
 /*         ||||||||   */
/* PB0 (5)  |||||||+-   Out        (output)  */
/* PB1 (6)  ||||||+--   Tx         (output)  */
/* PB2 (7)  |||||+---   Rx          (input)  */
/* PB3 (2)  ||||+----   NC         (output)  */
/* PB4 (3)  |||+-----   NC         (output)  */
/* PB5 (1)  ||+------   Tts         (input)  */ 
/*          ++-------   empty                */

PORTB =   0b00000011;
 /*         ||||||||   */
/* PB0 (5)  |||||||+-   Out        (output)  */
/* PB1 (6)  ||||||+--   Tx         (output)  */
/* PB2 (7)  |||||+---   Rx          (input)  */
/* PB3 (2)  ||||+----   NC         (output)  */
/* PB4 (3)  |||+-----   NC         (output)  */
/* PB5 (1)  ||+------   Tts         (input)  */ 
/*          ++-------   empty                */

// Crystal Oscillator division factor: 1 
CLKPR=0b10000000; 
CLKPR=0x00; 
CLKPR=0x00; 
CLKPR=0x00; 
CLKPR=0x00; 


TCCR0A = 0x00;
TCCR0B = 0x02; //Count clock/8 
TCNT0 = 121;// 256 - (7761760/8/7200)
 
/*-------------- Init ADC ------------------------*/
 

  ADCSRA = 0b10000110; //  - single mode run, no int, f/64
  ADCSRB = 0b00000000; //
  ADMUX=0b00100011; // vcc connected, left ajust, ADC3 (PB3) - 2
//  Start conversion and wait for it to complete (result not used)
  ADCSRA |= (1<<ADSC);            // Start convertion
  while (ADCSRA & (1<<ADSC));     // Wait for conversion to complete
  ADCSRA = ADCSRA;                // Clear ADC interrupt flag

/*------------------------------------------------*/
 /* Timer 0, peripheral int enable */
 TIMSK = 0B00000010; // enable timer 0 int
/*------------------------------------------------*/

/*----------- init tuart -------------------------*/
 rx_smpl_cntr = 0;
 rx_state = 0;

 tx_smpl_cntr = 0;
 tx_state = 0;

 tx_ok = 0;
 tidx = iidx = 0;
/*------------------------------------------------*/
 mOut = 0;
  tOn = eeprom_read_byte(&ee_tOn);
  tOff = eeprom_read_byte(&ee_tOff);
  sei(); 

 /* Main loop */
 for(;;)
   {
    //wdt_reset();
    /* Call state machine */ 
	    idle();
            term();

    if ((iidx == 0))
     {
      tputstr("\rT = "); 
      printi(Tts);
      if (mOut == 1) tputstr(" On ");
      else tputstr(" Off");
     }
   }
}

